import logging
from copy import copy
from numbers import Number
from typing import Generator, Iterable, Sequence

import numpy as np
import torch
import whisper

from live_transcribe.text import compute_delta, merge_texts


def rolling_transcribe(
        data_generator: Iterable[Sequence[Number]],
        model: str = "small",
        language: str = "en",
        audio_window_size: int = whisper.audio.SAMPLE_RATE * 5,
) -> Generator[str, None, None]:
    """
    Transcribe audio in a rolling fashion, i.e. the model is fed with the last `window_size` samples.
    :param data_generator: a generator that yields audio samples
    :param model: "tiny", "base", "small", "medium", "large"
    :param language: "en", "pl", ...
    :param audio_window_size: a number of samples to keep in the buffer (`whisper.audio.SAMPLE_RATE` samples per second)
    :return: a generator that yields transcriptions
    """
    if torch.cuda.is_available():
        DEVICE = "cuda"
    else:
        logging.warning("CUDA is not available. Using CPU instead. You will experience a significant latency.")
        DEVICE = "cpu"

    model_name = model + ".en" if language == "en" else model
    logging.info(f"Loading model {model_name}...")
    model = whisper.load_model(model_name, device=DEVICE)
    logging.debug(
        f"Model is {'multilingual' if model.is_multilingual else 'English-only'} "
        f"and has {sum(np.prod(p.shape) for p in model.parameters()):,} parameters."
    )
    logging.info("Model loaded.")

    options = whisper.DecodingOptions(
        language=language,
        suppress_tokens=[],
        suppress_blank=False,
        without_timestamps=False,
        fp16=False,
        length_penalty=0.1,
        # prompt=prev_text
    )

    prev_text = ""
    audio_samples = None
    for new_audio_data in data_generator:
        len_from_previous = audio_window_size - len(new_audio_data)
        if audio_samples is None or len_from_previous < 0:
            audio_samples = copy(new_audio_data)
        else:
            audio_samples = np.concatenate((audio_samples[-len_from_previous:], new_audio_data))

        audio_normalized = (audio_samples / np.max(audio_samples)).astype(np.float32)
        audio_padded = whisper.pad_or_trim(audio_normalized)
        mel = whisper.log_mel_spectrogram(audio_padded).to(model.device)
        result = whisper.decode(model, mel, options)

        curr_text = merge_texts(prev_text, result.text)
        yield compute_delta(prev_text, curr_text)
        prev_text = curr_text[-len(result.text):]



import threading
from typing import Mapping, Optional

import numpy as np
import pasimple


class PaSimpleCapture:
    NP_TO_PA_FORMAT: Mapping[np.dtype, int] = {
        np.int32: pasimple.PA_SAMPLE_S32LE,
        np.float32: pasimple.PA_SAMPLE_FLOAT32LE,
        np.int16: pasimple.PA_SAMPLE_S16LE,
    }

    def __init__(self, rate: int, device_name: str = None, format: np.dtype = np.int32):
        if format not in self.NP_TO_PA_FORMAT:
            raise ValueError(f"Unsupported format: {format}; supported formats: {self.NP_TO_PA_FORMAT.keys()}")
        pa_format = self.NP_TO_PA_FORMAT[format]
        self._pa = pasimple.PaSimple(pasimple.PA_STREAM_RECORD, pa_format, 1, rate, device_name=device_name,
                                     app_name="live_transcribe")
        self._pa2: Optional[pasimple] = None
        self._chunk_size_bytes: int = rate // 256 * pasimple.format2width(pa_format)  # 1/256 s of latency
        self._stop = False
        self._thread = threading.Thread(target=self._run)
        self._mutex = threading.Lock()
        self._data = bytearray()
        self._np_type = format
        self._exception: Optional[BaseException] = None

    def _run(self):
        try:
            while not self._stop:
                # print("recording...")
                new_data = self._pa2.read(self._chunk_size_bytes)
                # print(f"got {len(new_data)} bytes")
                with self._mutex:
                    self._data += new_data
        except Exception as e:
            self._exception = e

    def __enter__(self) -> 'PaSimpleCapture':
        self._pa2 = self._pa.__enter__()
        self._thread.start()
        return self

    def __exit__(self, exc_type, exc_val, exc_tb) -> bool:
        self._stop = True
        self._thread.join()
        return self._pa.__exit__(exc_type, exc_val, exc_tb)

    def __iter__(self):
        return self

    def __next__(self) -> np.ndarray:
        if self._exception is not None:
            raise self._exception
        with self._mutex:
            data = self._data
            self._data = bytearray()

        return np.frombuffer(data, dtype=self._np_type)



# create a Qt5 window and display a text in it

import sys
from PyQt5.QtWidgets import QApplication, QMainWindow, QWidget, QLabel
from PyQt5.QtGui import QIcon, QFont
from PyQt5.QtCore import pyqtSlot, Qt

app = QApplication([])
window = QMainWindow()
window.setWindowTitle('PyQt5 App')
# window.setGeometry(100, 100, 280, 80)
# window.move(60, 15)
label = QLabel('Hello \nWorld!', parent=window)
window.setCentralWidget(label)


window.show()
app.processEvents()
app.exec_()


import threading
from queue import Queue
from time import time

import numpy as np
import pasimple
import torch
import whisper
from colorama import Back, Cursor, Fore, Style
from whisper.tokenizer import get_tokenizer

from live_transcribe.text import merge_sequences, split_into_lines

MODEL = "small"
PA_FORMAT = pasimple.PA_SAMPLE_S32LE
NP_FORMAT = np.int32
SAMPLE_WIDTH = pasimple.format2width(PA_FORMAT)
CHANNELS = 1
SAMPLE_RATE = whisper.audio.SAMPLE_RATE
LENGTH = 0.1
LANGUAGE = "en"
MAX_LEN = SAMPLE_RATE * 5
FALSE = False

device_name = "bluez_output.00_1D_43_A0_C1_DF.a2dp-sink.monitor"
cookie = 0

if torch.cuda._is_compiled():
    DEVICE = "cuda"
else:
    print("CUDA is not available. Using CPU instead. You will experience a significant latency.")
    DEVICE = "cpu"

model_name = MODEL + ".en" if LANGUAGE == "en" else MODEL
model = whisper.load_model(model_name, device=DEVICE)
print(
    f"Model is {'multilingual' if model.is_multilingual else 'English-only'} "
    f"and has {sum(np.prod(p.shape) for p in model.parameters()):,} parameters."
)

chunks_queue = Queue()
cookie += 1


def record_loop():
    with pasimple.PaSimple(pasimple.PA_STREAM_RECORD, PA_FORMAT, CHANNELS, SAMPLE_RATE, device_name=device_name) as pa:
        my_cookie = cookie
        while cookie == my_cookie:
            # print("recording...")
            audio_data = pa.read(int(CHANNELS * SAMPLE_RATE * SAMPLE_WIDTH * LENGTH))
            # print("done...")
            chunks_queue.put((time(), audio_data))


rec_task = threading.Thread(target=record_loop)
rec_task.start()

timeout = LENGTH * 2

tokenizer = get_tokenizer(
    model.is_multilingual, language=LANGUAGE, task="transcribe"
)

best_lang = LANGUAGE

options = whisper.DecodingOptions(
    language=best_lang,
    suppress_tokens=[],
    suppress_blank=False,
    without_timestamps=False,
    fp16=False,
    length_penalty=0.1,
    # prompt=prev_text
)

MAX_NUM_LINES = 5
MAX_LINE_LEN = 80


# MAX_LEN = whisper.audio.N_SAMPLES
def trascribe_loop():
    audio_data = bytearray()
    my_cookie = cookie
    curr_text = ""
    prev_text = ""
    prev_end_ts = None
    while cookie == my_cookie:
        if FALSE:
            print(f"qsize: {chunks_queue.qsize()}")
        audio_data = bytearray(audio_data)
        end_ts, chunk = chunks_queue.get(timeout=timeout)
        audio_data += chunk
        while not chunks_queue.empty():
            end_ts, chunk = chunks_queue.get(timeout=timeout)
            audio_data += chunk

        audio_data = bytearray(audio_data[-MAX_LEN * SAMPLE_WIDTH:])
        audio = np.frombuffer(audio_data, dtype=NP_FORMAT)
        audio_normalized = (audio / np.max(audio)).astype(np.float32)

        audio_padded = whisper.pad_or_trim(audio_normalized)
        if FALSE:
            print(f"{len(audio_normalized)} -> {len(audio_padded)}")
        mel_cpu = whisper.log_mel_spectrogram(audio_padded)
        mel = mel_cpu.to(model.device)

        # _, probs = model.detect_language(mel)
        # best_lang = max(probs, key=probs.get)
        # print(f"Detected language: {best_lang}")

        # print(":::", end="")
        result = whisper.decode(model, mel, options)

        curr_text2 = " ".join(merge_sequences(curr_text.split(" "), result.text.split(" ")))
        if FALSE:
            print(curr_text)
            print(curr_text2)
            print(result.text)
        curr_text = curr_text2

        curr_text_lines = split_into_lines(curr_text, MAX_LINE_LEN)

        if len(curr_text_lines) > MAX_NUM_LINES:
            lines_to_remove = len(curr_text_lines) - MAX_NUM_LINES
            len_to_remove = sum(map(len, curr_text_lines[:lines_to_remove])) + lines_to_remove
            curr_text = curr_text[len_to_remove:]
            curr_text_lines = split_into_lines(curr_text, MAX_LINE_LEN)
            # sys.stderr.write(Cursor.DOWN(1))

        sys.stderr.write(Cursor.UP(len(curr_text_lines)) + "\n".join(curr_text_lines) + "\n")

        if FALSE:
            tokenizer.decode_with_timestamps(result.tokens)
            decoded_tokens, _ = tokenizer.split_to_word_tokens(result.tokens)
            print(prev_decoded_tokens)
            print(decoded_tokens)
            print(list(zip(prev_tokens, prev_decoded_tokens)))
            print(list(zip(result.tokens, decoded_tokens)))
            prev_end_ts
            end_ts
            print(prev_tokens)
            print(result.tokens)
            tokenizer.decode_with_timestamps(prev_tokens)
            tokenizer.decode_with_timestamps(result.tokens)

            prev_tokens = result.tokens
            prev_decoded_tokens = decoded_tokens
            prev_end_ts = end_ts
        # prev_text = result.text
    chunks_queue.task_done()


trans_task = threading.Thread(target=trascribe_loop)
trans_task.start()

import sys

print("11111111")
print("11111111")
sys.stdout.write(u"2222\u001b[5Deee")
print("333")

for i in range(0, 100):
    sys.stdout.write(u"\u001b[1000D" + str(i + 1) + "%")
    sys.stdout.flush()

import time, sys

print("\033[FMy text overwriting the previous line.")
print("\033[FMy text overwriting the previous line.")
print("\033[FMy text overwriting the previous line.")

import tqdm

for i in tqdm.tqdm(range(100)):
    time.sleep(0.1)

sys.stdout.write('aaa\x1b[Abbb')
sys.stdout.flush()


def move(y, x):
    print("\033[%d;%dH" % (y, x))


print("aaa")
move(-1, -1)
print("bbb")

from colorama import just_fix_windows_console, init

just_fix_windows_console()
init()

for i in range(5):
    sys.stderr.write(Fore.RED + 'some red text')
    sys.stderr.write(Fore.RED + 'some red text')
    sys.stderr.write(Fore.RED + 'some red text')
    sys.stderr.write(Cursor.UP(2) + Back.GREEN + 'and with a green background')
    sys.stderr.write(Style.DIM + 'and in dim text')
    sys.stderr.write(Style.RESET_ALL)
    sys.stderr.write('back to normal now')


def print_loop():
    for i in range(4):
        sys.stderr.write("aaaa\n")
        sys.stderr.write(Cursor.UP(2) + 'bbb\n')
        time.sleep(1)


rec_task = threading.Thread(target=record_loop)
rec_task.start()



from typing import Iterable, Sequence

import numpy as np


def merge_sequences(first_seq, second_seq):
    """
    Merge two overlapping strings.
    :param first_seq:
    :param second_seq:
    :return:

    >>> merge_sequences("one two three fourx ", "three four five six seven")
    'one two three four five six seven'
    >>> merge_sequences("aaa", "bbb")
    'aaabbb'
    >>> merge_sequences("aaa", "aa")
    'aaa'
    >>> merge_sequences("aa", "")
    'aa'
    >>> merge_sequences("", "aa")
    'aa'
    >>> merge_sequences("", "")
    ''
    >>> merge_sequences("aabbaa", "aaccaa")
    'aabbaaccaa'
    >>> merge_sequences(["aa", "bb"], ["bb", "cc"])
    ['aa', 'bb', 'cc']
    """
    DEC_FIRST = 0
    DEC_SECOND = 1
    DEC_NONE = 2
    DEC_BOTH = 3

    cost = np.zeros((len(first_seq) + 1, len(second_seq) + 1), dtype=np.int32)
    decision = np.zeros((len(first_seq) + 1, len(second_seq) + 1), dtype=np.int32)
    # cost[i, j] = cost of transforming earlier_seq[:i] into later_seq[:j]
    # later_seq should start inside earlier_seq, so cost for i>=0 and  j==0 is low
    # but cost for j>0 and i==0 is large
    # INF_COST = 1000
    REPLACE_COST = 4
    DELETE_MIDDLE_COST = 4
    UNPREFIX_FIRST_COST = 1
    UNPREFIX_SECOND_COST = 2
    UNSUFFIX_FIRST_COST = 2
    UNSUFFIX_SECOND_COST = 1
    cost[1:, 0] = UNPREFIX_FIRST_COST * np.arange(1, len(first_seq) + 1)
    decision[1:, 0] = DEC_FIRST
    cost[0, 1:] = UNPREFIX_SECOND_COST * np.arange(1, len(second_seq) + 1)
    decision[0, 1:] = DEC_SECOND

    for i in range(1, len(first_seq) + 1):
        for j in range(1, len(second_seq) + 1):
            candidates = [
                (cost[i - 1, j - 1] + REPLACE_COST, DEC_NONE),
                (cost[i - 1, j] + (DELETE_MIDDLE_COST if j < len(second_seq) else UNSUFFIX_FIRST_COST), DEC_FIRST),
                (cost[i, j - 1] + (DELETE_MIDDLE_COST if i < len(first_seq) else UNSUFFIX_SECOND_COST), DEC_SECOND),
            ]
            if first_seq[i - 1] == second_seq[j - 1]:
                cost[i, j] = cost[i - 1, j - 1]
                decision[i, j] = DEC_BOTH
                candidates.append((cost[i - 1, j - 1], DEC_BOTH))

            # if i == len(first_seq):
            #     print(f"i={i}, j={j}, candidates={candidates}")
            cost[i, j], decision[i, j] = min(candidates, key=lambda x: x[0])


    def backtrack(i, j):
        if i == 0 and j == 0:
            return None
        else:
            d = decision[i, j]
            if d == DEC_FIRST:
                prev = backtrack(i - 1, j)
                letter = first_seq[i - 1:i] if j == 0 else None
            elif d == DEC_SECOND:
                prev = backtrack(i, j - 1)
                letter = second_seq[j - 1:j] if i == len(first_seq) else None
            elif d == DEC_BOTH:
                prev = backtrack(i - 1, j - 1)
                letter = first_seq[i - 1:i]
            else:
                prev = backtrack(i - 1, j - 1)
                if j < len(second_seq) - i:  # near ends there are greater risk of mistakes
                    letter = first_seq[i - 1:i]
                else:
                    letter = second_seq[j - 1:j]
            # print(f"{letter} {i} {j} {d} {cost[i, j]}")

            if letter is None:
                return prev
            elif prev is None:
                return letter
            else:
                return prev + letter

    res = backtrack(len(first_seq), len(second_seq))
    if res is None:
        res = type(first_seq)()
    return res


def merge_texts(text1: str, text2: str) -> str:
    DELIMITER = " "
    return DELIMITER.join(merge_sequences(text1.split(DELIMITER), text2.split(DELIMITER)))


def common_prefix_len(a: Sequence, b: Sequence) -> int:
    for i, (x, y) in enumerate(zip(a, b)):
        if x != y:
            return i
    return min(len(a), len(b))


Delta = str  # a string to append that may contain backspaces
BACKSPACE_CHARACTER: str = "\x08"


def compute_delta(base: str, new: str) -> Delta:
    """
    >>> compute_delta("hello", "hell")
    '\\x08'
    >>> compute_delta("hello", "hello world")
    ' world'
    >>> compute_delta("hello xyz", "hello world")
    '\\x08\\x08\\x08 world'
    >>> compute_delta("hello", "hello")
    ''
    """
    cpl = common_prefix_len(base, new)
    return BACKSPACE_CHARACTER * (len(base) - cpl) + new[cpl:]


def apply_delta(s1: str, diff: Delta) -> str:
    """
    >>> apply_delta("hello", "\\x08")
    'hell'
    >>> apply_delta("hello", " world")
    'hello world'
    >>> apply_delta("hello xyz", "\\x08\\x08\\x08world")
    'hello world'
    >>> apply_delta("hello", "")
    'hello'
    """
    res = s1
    for c in diff:
        if c == BACKSPACE_CHARACTER:
            res = res[:-1]
        else:
            res += c
    return res


def split_into_lines(text: str, max_line_length: int):
    """
    Split text into lines of max_line_length characters.

    :param text:
    :param max_line_length:
    :return:

    >>> split_into_lines("one two three four five six seven", 10)
    ['one two', 'three four', 'five six', 'seven']
    >>> split_into_lines("123456789012 1 123456789012 32 43", 10)
    ['123456789012', '1', '123456789012', '32 43']


    """
    lines = []
    line = ""
    for word in text.split(" "):
        if len(line) + len(word) > max_line_length and line:
            lines.append(line)
            line = word
        else:
            if line:
                line += " "
            line += word
    if line:
        lines.append(line)
    return lines


def split_into_lines_live(deltas_generator: Iterable[Delta], preferred_line_length: int, max_line_length: int):
    """
    When the last line is longer than max_line_length, split out the "preffered_line_length" part of it to another line.
    >>>  list(split_into_lines_live(["one", " two", " three", " four", " five", " six", " seven"], 5, 10))
    """
    text = ""
    for delta in deltas_generator:
        prev_text = text
        text = apply_delta(text, delta)
        # print(f"delta={delta}, text={text}")
        if len(text) < max_line_length:
            yield delta
        else:
            lines = split_into_lines(text, preferred_line_length)
            out_delta = compute_delta(prev_text, "\n".join(lines))
            text = lines[-1]
            yield out_delta


list(split_into_lines_live(["one", " two", " three", " four", " five", " six", " seven"], 5, 10))



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['live_transcribe']

package_data = \
{'': ['*']}

install_requires = \
['numpy>=1.23,<2.0,!=1.24.0,!=1.24.1,!=1.24.2',
 'openai-whisper>=20230314,<20230315',
 'pasimple>=0.0.1,<0.0.2',
 'pulsectl>=22.3.2,<23.0.0']

entry_points = \
{'console_scripts': ['live-transcribe = live_transcribe:__main__']}

setup_kwargs = {
    'name': 'live-transcribe',
    'version': '0.1.0',
    'description': "A live, real-time transcription of the audio based on OpenAI's Whisper API.",
    'long_description': "Live Transcribe\n\nLive Transcribe is a Python package that provides live, real-time transcription of audio based on OpenAI's Whisper API.\n\nCurrently, Live Transcribe supports only PulseAudio as an audio backend.\n\n# Installation\n\n1. **optional but highly recommended for low latency** Refer to the [Pytorch documentation guide](https://pytorch.org/)\n   to install Pytorch **with CUDA support**.\n1. `pip install live-transcribe`\n\n# Usage\n\nJust run:\n\n    python -m live_transcribe\n\nOn the first usage, the OpenAI's Whisper model will be downloaded and cached.\n\nSee `live_transcribe --help` for options.\n\n# Dependencies\n\nLive Transcribe has the following dependencies:\n\n    Python 3.8 or higher\n    OpenAI-Whisper\n    PulseAudio\n",
    'author': 'Tomasz Łakota',
    'author_email': 'tomasz.lakota@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.8,<4.0',
}


setup(**setup_kwargs)



import argparse
import logging
import sys
from contextlib import suppress

import pulsectl
import whisper

from live_transcribe.audio_pasimple import PaSimpleCapture
from live_transcribe.transcribe import rolling_transcribe

logging.basicConfig(level=logging.INFO)


def obtain_default_device_name():
    pulse = pulsectl.Pulse('live_transcribe')
    sinks_by_name = {i.name: i for i in pulse.sink_list()}
    default_sink = sinks_by_name[pulse.server_info().default_sink_name]
    return default_sink.monitor_source_name


def print_devices():
    pulse = pulsectl.Pulse('live_transcribe')
    for i in pulse.source_list():
        with suppress(AttributeError):
            print(f"  {i.name}: {i.description}")


# program parameters
argparser = argparse.ArgumentParser()
argparser.add_argument("--device", help="audio device name; by default the monitor of the default sink is used")
argparser.add_argument("--model", help="model name", default="small")
argparser.add_argument("--language", help="language", default="en")
argparser.add_argument("--window",
                       type=float,
                       help="a length (in seconds) of audio sample passed each time to the model; greater values improve quality but increase latency",
                       default=4)
argparser.add_argument("--list-devices", action="store_true", help="list available audio devices and exit")

args = argparser.parse_args()

if args.list_devices:
    print_devices()
    sys.exit(0)

audio_device_name = args.device or obtain_default_device_name()
logging.info(f"Using audio device {audio_device_name}")
stream = sys.stdout

rate = whisper.audio.SAMPLE_RATE
with PaSimpleCapture(rate, device_name=audio_device_name) as pa:
    # for txt in  split_into_lines_live(rolling_transcribe(pa), 80, 100):
    for txt in rolling_transcribe(pa,
                                  model=args.model,
                                  language=args.language,
                                  audio_window_size=int(args.window * rate)):
        stream.write(txt)
        stream.flush()
